from pathlib import Path
from typing import Any, Callable, Iterable, Mapping, Optional, Union

import sqlparse
from django.conf import settings
from django.db import DEFAULT_DB_ALIAS, connections
from django.test import TestCase
from django.test.utils import CaptureQueriesContext
from django.utils.functional import cached_property
from rest_framework.test import APITestCase

from arkindex.documents.models import Corpus
from arkindex.images.models import ImageServer
from arkindex.users.models import User


class _AssertExactQueriesContext(CaptureQueriesContext):
    """
    Context manager for assertExactQueries.
    The implementation is inspired by assertNumQueries's own implementation.
    """

    def __init__(self, test_case, path, params, connection):
        self.test_case = test_case
        self.path = settings.SQL_VALIDATION_DIR / Path(path)
        self.params = params
        super().__init__(connection)

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        if exc_type is not None:
            return

        # Django's logged queries are each on a single line without semicolons
        # so we lint them and assemble them to build a more readable diff for humans.
        actual_sql = sqlparse.format(
            ';'.join(query['sql'] for query in self),
            reindent=True,
            use_space_around_operators=True,
            indent_width=4,
        )

        # When the file does not exist, try to create it with the current queries.
        # This avoids having to write all of the queries by hand when adding assertExactQueries for the first time.
        if not self.path.exists():
            try:
                self.path.write_text(actual_sql)
            except IOError as e:
                raise AssertionError(
                    f'Could not assert on SQL queries; the file at {self.path} did not exist. '
                    'A new file could not be created with the current SQL queries.'
                ) from e
            else:
                raise AssertionError(
                    f'Could not assert on SQL queries; the file at {self.path} did not exist. '
                    'A file has been created with the current SQL queries, please check it manually.'
                )

        expected_sql = self.path.read_text().strip()
        # Apply str.format on the SQL when needed to handle variable IDs
        if self.params:
            if isinstance(self.params, Mapping):
                expected_sql = expected_sql.format(**self.params)
            else:
                expected_sql = expected_sql.format(*self.params)

        # Use the base TestCase's assertions, since they can already handle displaying a diff.
        self.test_case.assertEqual(expected_sql, actual_sql)


class FixtureMixin(object):
    """
    Add the database fixture to a test case
    """

    fixtures = ['data.json', ]

    @classmethod
    def setUpTestData(cls):
        cls.corpus = Corpus.objects.get(name='Unit Tests')
        cls.user = User.objects.get(email='user@user.fr')
        cls.superuser = User.objects.get(email='root@root.fr')
        cls.imgsrv = ImageServer.objects.get(url='http://server')

    @cached_property
    def internal_user(self):
        return User.objects.get(email='internal@internal.fr')

    def assertExactQueries(
            self,
            path: Union[Path, str],
            func: Optional[Callable] = None,
            *args,
            using: str = DEFAULT_DB_ALIAS,
            params: Union[Iterable, Mapping[str, Any]] = [],
            **kwargs) -> Optional[_AssertExactQueriesContext]:
        """
        Assert that a function call causes exactly the SQL queries specified in a given file.
        When used as a context manager, asserts than any queries occurring while in the context match the given file.
        This function supports all of Django's `assertNumQueries` syntax:

        >>> self.assertExactQueries('path/to/file.sql', do_something)
        >>> with self.assertExactQueries('path/to/file.sql'):
        ...     do_something()

        Paths will be relative to the SQL_VALIDATION_DIR setting.
        SQL queries can often hold IDs or other dynamic values that depend on the test context.
        To avoid having to update IDs or rewrite the fixture file every time, you can use the `params` argument
        and give it anything supported by `str.format`; either an iterable or a mapping. For example:

        ```
        SELECT "documents_element"."name" FROM "documents_element" WHERE "documents_element"."id" = '{id}'::uuid;
        ```

        >>> with self.assertExactQueries('path/to/file.sql', params={'id': element_id}):
        ...     Element.objects.filter(id=element_id).values('name')

        To make first runs easier, when the expect SQL file does not exist, assertExactQueries will try to write
        the current SQL queries to it and warn you using an `AssertionError`. Make sure to check this new SQL file
        as subsequent runs of unit tests will no longer fail.
        """
        context = _AssertExactQueriesContext(self, path, params, connections[using])

        if func is None:
            return context

        with context:
            func(*args, **kwargs)


class FixtureTestCase(FixtureMixin, TestCase):
    """
    Django test case with the database fixture
    """


class FixtureAPITestCase(FixtureMixin, APITestCase):
    """
    Django REST Framework test case with the database fixture
    """
